\ 2017-03-27 EW   ewlib/clock_tick0_external.fs
\
\ Written in 2017 by Erich Wälde <erich.waelde@forth-ev.de>
\
\ To the extent possible under law, the author(s) have dedicated
\ all copyright and related and neighboring rights to this software
\ to the public domain worldwide. This software is distributed
\ without any warranty.
\
\ You should have received a copy of the CC0 Public Domain
\ Dedication along with this software. If not, see
\ <http://creativecommons.org/publicdomain/zero/1.0/>.
\
\ words:
\        +ticks    register and enable interupt
\        -ticks    disable interupt

#128 constant ticks/sec
variable ct.ticks
variable ct.ticks.follow
variable last.tick[6]
variable last.tick[7]

\ ct.ticks is used as 8-bit counter (not 16-bit)
: c0<        ( -- )     >< 0< ;
: tick.over? ( -- t/f ) ct.ticks.follow c@  ct.ticks c@  - c0< ;
: tick.over! ( -- )     ct.ticks.follow c@  1+  ct.ticks.follow c! ;

: tick_isr
  ct.ticks c@ 1+  ct.ticks c!
;

\ enable ticks
\ ext.clock: 32768 /sec
\ overflow:  32768/256 = 128 /sec =^= 7.8125 milli-sec ticks
: +ticks
  0 ct.ticks         !
  0 ct.ticks.follow  !
  0 last.tick[6]     !
  0 last.tick[7]     !
  
  [
    %00000000                           \ OC.A,B off, normal mode
  ] literal TCCR0A c! 
  [
    %00000110                           \ ext. clocksource on T0, falling edge
  ] literal TCCR0B c!

  \ pin T0 input, low = highZ?
  DDRA  c@ $01 invert and DDRA  c!      \ pin T0 input
  PORTA c@ $01 invert and PORTA c!      \ + low  => highZ

  ['] tick_isr TIMER0_OVFAddr int!      \ register interupt service routine
  
  TIMSK0 c@ $01 or TIMSK0 c!            \ timer0 overflow int. enable
;

\ disable ticks
: -ticks
  TIMSK0 c@ $01 invert and TIMSK0 c!    \ disable timer0 ovf int
  $00  TCCR0B c!
  $07  TIFR0  c!                        \ clear interrupt flags, jibc
;

\ one second == 128 ticks
\ half second == 64 ticks
\ that is a toggle on bit 6 of ct.ticks.follow
: half.second.over? ( -- 0|1|2 )
  \ return: 0 == false
  \         1 == half second over
  \         2 == second over
  ct.ticks.follow c@
  $0040 and 0= 0=  \ extract significant bit as t/f
  dup last.tick[6] @ = if
    \ no change, done
    drop 0
  else
    dup 0= if
      \ falling edge, second over
      2
    else
      \ rising edge, half second over
      1
    then
    swap
    ( sig.bit-t/f ) last.tick[6] !
  then
;

: second.over? ( -- t/f )
  ct.ticks.follow c@  $0080 and 0= 0=
  dup  last.tick[7] @  = if
    drop 0
  else
    last.tick[7] !
    -1
  then
;
