\ 2017-08-13 ewlib/clock_tick1_main.fs
\
\ Written in 2017 by Erich Wälde <erich.waelde@forth-ev.de>
\
\ To the extent possible under law, the author(s) have dedicated
\ all copyright and related and neighboring rights to this software
\ to the public domain worldwide. This software is distributed
\ without any warranty.
\
\ You should have received a copy of the CC0 Public Domain
\ Dedication along with this software. If not, see
\ <http://creativecommons.org/publicdomain/zero/1.0/>.
\

#32 constant ticks/sec
variable ct.ticks
variable ct.ticks.follow
variable last.tick[4]
variable last.tick[5]

\ overflow interupt service routine
: tick_isr
  1 ct.ticks +!
;

: tick.over? ( -- t/f ) ct.ticks.follow @  ct.ticks @  - 0< ;
: tick.over! ( -- )     1 ct.ticks.follow +! ;

\ enable ticks
\ crystal:   11059200 /sec
\ prescaler: 256
\            43200 /sec
\ TOP+1:     1350
\            32 /sec
: +ticks
  0 ct.ticks        !
  0 ct.ticks.follow !
  0 last.tick[4]    !
  0 last.tick[5]    !

  \ --- timer1 ! ---
  [ %00000000                 \ WGM1[10] CTC mode
    %01000000 or              \ COM1A[10] toggle OC1A on compare match
  ] literal TCCR1A c!
  #1350 1-  OCR1A   !         \ TOP or compare match value rather
  DDRD c@ $80 or DDRD c!      \ pin OC2A output  
  [ %00000100                 \ CS1[210] clock_ts2/256
    %00001000 or              \ WGM1[32] CTC mode
  ] literal TCCR1B c! 
                              \ register isr
  ['] tick_isr TIMER1_COMPAAddr int! 
  TIMSK1 c@ $02 or TIMSK1 c!  \ enable OCIE1A interupt
;

: -ticks
  TIMSK1 c@
  [ $02 invert ] literal
  and TIMSK1  c!              \ disable OCIE1A interrupt
  $00  TCCR1B c!              \ disable timer/counter1
  $02  TIFR2  c!              \ clear interrupt flags
;

\ no phase shift accumulator
\ one second == 32 ticks
\ half second == 16 ticks
\ that is a toggle on bit 4 of ct.ticks.follow
: half.second.over? ( -- 0|1|2 )
  \ return: 0 == false
  \         1 == half second over
  \         2 == second over
  ct.ticks.follow c@
  $0010 and 0= 0=  \ extract significant bit as t/f
  dup last.tick[4] @ = if
    \ no change, done
    drop 0
  else
    dup 0= if
      \ falling edge, second over
      2
    else
      \ rising edge, half second over
      1
    then
    swap
    ( sig.bit-t/f ) last.tick[4] !
  then
;

: second.over? ( -- t/f )
  ct.ticks.follow c@  $0020 and 0= 0=
  dup  last.tick[5] @  = if
    drop 0
  else
    last.tick[5] !
    -1
  then
;
