\ 2017-03-27 EW   ewlib/clock_tick2_clockcrystal.fs
\
\ Written in 2017 by Erich Wälde <erich.waelde@forth-ev.de>
\
\ To the extent possible under law, the author(s) have dedicated
\ all copyright and related and neighboring rights to this software
\ to the public domain worldwide. This software is distributed
\ without any warranty.
\
\ You should have received a copy of the CC0 Public Domain
\ Dedication along with this software. If not, see
\ <http://creativecommons.org/publicdomain/zero/1.0/>.
\

#128 constant ticks/sec
variable ct.ticks
variable ct.ticks.follow
variable last.tick[6]
variable last.tick[7]


\ timer 2 overflow interrupt service routine
: tick2_isr
  1 ct.ticks +!
;

: tick.over? ( -- t/f ) ct.ticks.follow @  ct.ticks @  - 0< ;
: tick.over! ( -- )     1 ct.ticks.follow +! ;


\ enable ticks
\ crystal:   32768 /sec
\ clock src: 32768 /sec
\ overflow:  32768/256 = 128 /sec
: +ticks
  0 ct.ticks         !
  0 ct.ticks.follow  !
  0 last.tick[6]     !
  0 last.tick[7]     !

  \ --- timer2 ---
  [ %00000000                   \ normal mode
  ] literal TCCR2A c!
  [ %00000001                   \ 001 = clock_ts2
  ] literal TCCR2B c! 
  ASSR_AS2 ASSR  c!             \ source: 32 kiHz crystal 
  ['] tick2_isr TIMER2_OVFAddr int! \ register ISR
  TIMSK2 c@ $01 or TIMSK2 c!    \  enable OVF2 interupt
;

\ disable ticks
: -ticks
  TIMSK2 c@
  [ $01 invert ] literal
  and TIMSK2  c! ( clr Timer 2 )
  $00  ASSR   c!
  $00  TCCR2B c!
  $07  TIFR2  c! \ clear interrupt flags, jibc
;


\ one second == 128 ticks
\ half second == 64 ticks
\ that is a toggle on bit 6 of ct.ticks.follow
: half.second.over? ( -- 0|1|2 )
  \ return: 0 == false
  \         1 == half second over
  \         2 == second over
  ct.ticks.follow c@
  $0040 and 0= 0=  \ extract significant bit as t/f
  dup last.tick[6] @ = if
    \ no change, done
    drop 0
  else
    dup 0= if
      \ falling edge, second over
      2
    else
      \ rising edge, half second over
      1
    then
    swap
    ( sig.bit-t/f ) last.tick[6] !
  then
;

: second.over? ( -- t/f )
  ct.ticks.follow c@  $0080 and 0= 0=
  dup  last.tick[7] @  = if
    drop 0
  else
    last.tick[7] !
    -1
  then
;
