\ 2017-07-23  dcf_01.fs
\
\ Written in 2017 by Erich Wälde <erich.waelde@forth-ev.de>
\
\ To the extent possible under law, the author(s) have dedicated
\ all copyright and related and neighboring rights to this software
\ to the public domain worldwide. This software is distributed
\ without any warranty.
\
\ You should have received a copy of the CC0 Public Domain
\ Dedication along with this software. If not, see
\ <http://creativecommons.org/publicdomain/zero/1.0/>.
\
\ functions to sample dcf signal
\     extract bit values,
\     extract values from bits,
\     verify parity
\     provide debug output
\
\ needs
\     #2000 constant Century
\     PORTD 7 portpin: _dcf
\     PORTB 2 portpin: led.0
\     : led_dcf  [: led.0 ;] execute ;
\
\ words
\     +dcf -dcf
\     f.dcf.insync
\     f.dcf.commit
\     +DD +DRX

variable dcfFlags
dcfFlags   0 flag: f.dcf.err            \ any error, telegram invalid
\  Flags   1 flag: f.dcf.???            \
dcfFlags  #2 flag: f.dcf.commit         \ commit dcf -> masterclock wanted
dcfFlags  #3 flag: f.dcf.dbg            \ debug output active
dcfFlags  #4 flag: f.dcf.nosignal       \ 
dcfFlags  #5 flag: f.dcf.insync         \ tell the world, dcf time is ok
dcfFlags  #6 flag: f.dcf.CET            \ timezone bit
dcfFlags  #7 flag: f.dcf.CEDT           \ daylight saving time bit
dcfFlags  #8 flag: f.dcf.parerr.min     \ parity error minute
dcfFlags  #9 flag: f.dcf.parerr.hour    \ parity error hour
dcfFlags #10 flag: f.dcf.parerr.date    \ parity error date
dcfFlags #11 flag: f.dcf.parerr         \ any parity error
dcfFlags #12 flag: f.dcf.dbg.rx         \ debug output per received bit active
\  Flags #13 flag: f.dcf.???
\  Flags #14 flag: f.dcf.???
\  Flags #15 flag: f.dcf.???

variable dcfPulse                       \ sample count, reduced tx power
: dcfPulse.clr  ( -- )  0 dcfPulse  ! ;
: dcfPulse++    ( -- )  1 dcfPulse +! ;
variable dcfBit                         \ value of last received bit 
variable dcfTick                        \ separate tick for dcf
: dcfTick++/mod ( -- )  dcfTick @ 1+ ticks/sec mod dcfTick ! ;
variable dcfEdge                        \ edge detector
#4                               constant dcfEdge:leading.bits
1 dcfEdge:leading.bits lshift 1- constant dcfEdge:leading.mask
: dcfEdge<<1    ( -- )  dcfEdge @ 1 lshift dcfEdge ! ;
: dcfEdge+1     ( -- )  1 dcfEdge +! ;
: dcfEdge?      ( -- )  dcfEdge @ dcfEdge:leading.mask = ;

: dcfTick.set   ( -- )  dcfEdge:leading.bits dcfTick ! ;

variable dcfPos                         \ position in telegram, == dcf second
: dcfPos++      ( -- )  dcfPos @ 1+ #60 mod dcfPos ! ;
variable dcfPar                         \ collect parity
: dcfPar++      ( -- )  1 dcfPar +! ;
: dcfPar?       ( -- t/f )  dcfPar @ $01 and 0= ; \ even parity
variable dcfErrCnt                      \ counter (down) of passed checks
: dcfErrCnt--   ( -- )  -1 dcfErrCnt +! ;
: dcfErrCnt.set ( -- )  #7 dcfErrCnt  ! ;

\ these collect the bits in the DCF bit stream
variable dcfTMin
variable dcfTHour
variable dcfTDay
variable dcfTWday
variable dcfTMonth
variable dcfTYear

: +DD  f.dcf.dbg fset ;
: -DD  f.dcf.dbg fclr ;
: .DD
  ." dcf.error:  " f.dcf.err     fset? if 1 else 0 then . cr
  ." dcf.parity: " f.dcf.parerr  fset? if 1 else 0 then . cr
  ." dcf.commit: " f.dcf.commit  fset? if 1 else 0 then . cr
  ." dcf.debug:  " f.dcf.dbg     fset? if 1 else 0 then . cr
  ." dcf.dbg.rx: " f.dcf.dbg.rx  fset? if 1 else 0 then . cr
;
: +DRX  f.dcf.dbg.rx fset ;
: -DRX  f.dcf.dbg.rx fclr ;


\ create temporary RAM table
variable dcf.tmp.table #60 cells allot \  only temporary really!
\ fill RAM table with ' drop  \ noop -> drop: remove argument!
' drop dcf.tmp.table #60 ramtable.init
\ dcf.tmp.table #60 ramtable.dump \ DEBUG
: >rt ( xt idx -- )  dcf.tmp.table >ramtable ;

\ code snippet XTs to ram table
\ structure of functions
\ :noname  ( 0|1 -- )
\     if   ( bit:1  ) ...
\     else ( bit:0  ) ...
\     then ( always ) ...
\ ;                                                      #idx >rt
\ :noname 
\     drop ( always ) ...
\ ;                                                      #idx >rt

\ #0   always 0!
:noname
  0= if ( bit:0 ) dcfErrCnt--
  then
;                                                           #0 >rt
\ #1 .. #14  civil warning bits, MeteoTime data bits
\ #15  call bit
\ #16  1: day light savings time switch at the end of this hour
\ #17  0 CET, 1 CEST
:noname
  if   ( bit:1 ) f.dcf.CEDT fset
  else ( bit:0 ) f.dcf.CEDT fclr
  then
;                                                           #17 >rt
\ #18  1 CET, 0 CEST
:noname
  if   ( bit:1 ) f.dcf.CET fset   f.dcf.CEDT fclr? if dcfErrCnt-- then
  else ( bit:0 ) f.dcf.CET fclr   f.dcf.CEDT fset? if dcfErrCnt-- then
  then
;                                                           #18 >rt
\ #19  1: leap second announcement
\ #20  always 1: start of encoded time
:noname
    if   ( bit:1 ) dcfErrCnt--
    then 
    f.dcf.parerr      fset
    f.dcf.parerr.min  fset
    f.dcf.parerr.hour fset
    f.dcf.parerr.date fset
    0 dcfPar !
;                                                           #20 >rt
\ #21 .. #28: minute
\ minute ones
:noname if ( bit:1 )   #1 dcfTMin   +!  dcfPar++ then ;     #21 >rt
:noname if ( bit:1 )   #2 dcfTMin   +!  dcfPar++ then ;     #22 >rt
:noname if ( bit:1 )   #4 dcfTMin   +!  dcfPar++ then ;     #23 >rt
:noname if ( bit:1 )   #8 dcfTMin   +!  dcfPar++ then ;     #24 >rt
\ minute tens
:noname if ( bit:1 )  #10 dcfTMin   +!  dcfPar++ then ;     #25 >rt
:noname if ( bit:1 )  #20 dcfTMin   +!  dcfPar++ then ;     #26 >rt
:noname if ( bit:1 )  #40 dcfTMin   +!  dcfPar++ then ;     #27 >rt
\ minute parity bit 
:noname
  if   ( bit:1 )                       dcfPar++ then 
  ( always )
  dcfPar? if f.dcf.parerr.min fclr  dcfErrCnt-- then
  0 dcfPar !
;                                                           #28 >rt

\ #29 .. #35: hour
\ hour ones
:noname if ( bit:1 )   #1 dcfTHour  +!  dcfPar++ then ;     #29 >rt
:noname if ( bit:1 )   #2 dcfTHour  +!  dcfPar++ then ;     #30 >rt
:noname if ( bit:1 )   #4 dcfTHour  +!  dcfPar++ then ;     #31 >rt
:noname if ( bit:1 )   #8 dcfTHour  +!  dcfPar++ then ;     #32 >rt
\ hour tens
:noname if ( bit:1 )  #10 dcfTHour  +!  dcfPar++ then ;     #33 >rt
:noname if ( bit:1 )  #20 dcfTHour  +!  dcfPar++ then ;     #34 >rt
\ hour parity bit
:noname
  if   ( bit:1 )                        dcfPar++  then
  ( always )
  dcfPar? if f.dcf.parerr.hour fclr  dcfErrCnt-- then
  0 dcfPar !
;                                                           #35 >rt
\ #36 .. #41: day
\ day ones
:noname if ( bit:1 )   #1 dcfTDay   +!  dcfPar++ then ;     #36 >rt
:noname if ( bit:1 )   #2 dcfTDay   +!  dcfPar++ then ;     #37 >rt
:noname if ( bit:1 )   #4 dcfTDay   +!  dcfPar++ then ;     #38 >rt
:noname if ( bit:1 )   #8 dcfTDay   +!  dcfPar++ then ;     #39 >rt
\ day tens
:noname if ( bit:1 )  #10 dcfTDay   +!  dcfPar++ then ;     #40 >rt
:noname if ( bit:1 )  #20 dcfTDay   +!  dcfPar++ then ;     #41 >rt
\ #42 .. #44: Wochentag
\ day of week
:noname if ( bit:1 )   #1 dcfTWday  +!  dcfPar++ then ;     #42 >rt
:noname if ( bit:1 )   #2 dcfTWday  +!  dcfPar++ then ;     #43 >rt
:noname if ( bit:1 )   #4 dcfTWday  +!  dcfPar++ then ;     #44 >rt
\ #45 .. #49: Monat
\ month ones
:noname if ( bit:1 )   #1 dcfTMonth +!  dcfPar++ then ;     #45 >rt
:noname if ( bit:1 )   #2 dcfTMonth +!  dcfPar++ then ;     #46 >rt
:noname if ( bit:1 )   #4 dcfTMonth +!  dcfPar++ then ;     #47 >rt
:noname if ( bit:1 )   #8 dcfTMonth +!  dcfPar++ then ;     #48 >rt
\ month tens
:noname if ( bit:1 )  #10 dcfTMonth +!  dcfPar++ then ;     #49 >rt
\ #50 .. #57: Jahr % 100
\ year ones
:noname if ( bit:1 )   #1 dcfTYear  +!  dcfPar++ then ;     #50 >rt
:noname if ( bit:1 )   #2 dcfTYear  +!  dcfPar++ then ;     #51 >rt
:noname if ( bit:1 )   #4 dcfTYear  +!  dcfPar++ then ;     #52 >rt
:noname if ( bit:1 )   #8 dcfTYear  +!  dcfPar++ then ;     #53 >rt
\ year tens
:noname if ( bit:1 )  #10 dcfTYear  +!  dcfPar++ then ;     #54 >rt
:noname if ( bit:1 )  #20 dcfTYear  +!  dcfPar++ then ;     #55 >rt
:noname if ( bit:1 )  #40 dcfTYear  +!  dcfPar++ then ;     #56 >rt
:noname if ( bit:1 )  #80 dcfTYear  +!  dcfPar++ then ;     #57 >rt
\ date parity bit
:noname
  if   ( bit:1 )                        dcfPar++ then
  ( always )
  dcfPar? if f.dcf.parerr.date fclr  dcfErrCnt-- then
  f.dcf.parerr.min  fclr? 
  f.dcf.parerr.hour fclr? and
  f.dcf.parerr.date fclr? and if f.dcf.parerr fclr then
;                                                           #58 >rt


\ dcf.tmp.table #60 ramtable.dump \ DEBUG
\ copy RAM table to FLASH
dcf.tmp.table #60 >flashtable pos_cmd_map
\ release RAM
dcf.tmp.table to here \ fixme: possibly bad???

: pos.cmd ( index -- )
  dup 0 #60 within if
    ( position ) pos_cmd_map +  @i execute
  else
    drop
  then
;

-2 constant dcf.bit:sync59
-1 constant dcf.bit:error
 0 constant dcf.bit:0
 1 constant dcf.bit:1
: dcf.pulse>bit ( pulse -- bit|error )
  dup  #2 < if                dcf.bit:sync59 else
  dup  #9 < if f.dcf.err fset dcf.bit:error  else
  dup #13 < if                dcf.bit:0      else
  dup #20 < if f.dcf.err fset dcf.bit:error  else
  dup #26 < if                dcf.bit:1      else
               f.dcf.err fset dcf.bit:error
  then then then then then
  swap drop
;

: dcf.led ( t/f -- )
  if led_dcf on else led_dcf off then
;
: dcf.dbg.minute.over ( -- )
  space
  f.dcf.err fclr? if
    [char] O emit
  else
    [char] F emit
  then
  cr
;
: dcf.dbg.second.over ( -- )
  f.dcf.dbg.rx fset? if
    dcfBit @ case
      dcf.bit:0      of [char] . emit endof
      dcf.bit:1      of [char] + emit endof
      dcf.bit:error  of [char] / emit endof
      dcf.bit:sync59 of [char] S emit endof
    endcase
  then
;



\ --- dcf.tick: 128/s, sample signal  ----
Rdefer dcf.tick
Rdefer dcf.sec
Rdefer dcf.min

: (dcf.tick)
    dcfClock  1 tick +!         \ keep dcfClock ticking
    dcfTick++/mod               \ .
    dcfEdge<<1                  \ . edge detection

    _dcf pin_low? if
        \ active!
        dcfPulse++
        dcfEdge+1
        -1 dcf.led
    else
        0 dcf.led
    then

    dcfEdge? if                 \ keep dcfClock in sync with radio signal
        dcfTick.set             \ . this might be bad occasionally
    then

    \ dcfTick: 0 .. 127
    \ 127 indicates the last tick in this second
    dcfTick @ ticks/sec 1- = if
        dcf.sec                 \ handle "one second over"
    then

    \ it is unclear whether deriving the dcf-second is
    \ ok from dcfTick, or whether ist must be
    \ dcfClock tick
    \ (dcf.tick) does both increments
    \ dcfTick is clear when calling dcf.sec
    \ dcf.sec clears dcfClock tick
;

\ --- dcf.sec: 1/s, book keeping ---------
: dcf.tmp.counter.reset
  #59 dcfPos  !
  0 dcfTMin   !
  0 dcfTHour  !
  0 dcfTDay   !
  0 dcfTWday  !
  0 dcfTMonth !
  0 dcfTYear  !
  0 dcfPar    !
;  
: dcfTemp>dcfClock
  dcfClock
  0                       tick  !
  0                       sec   !
  dcfTMin   @             min   !
  dcfTHour  @             hour  !
  dcfTDay   @  1-         day   !
\ dcfTWday  @             ?
  dcfTMonth @  1-         month !
  dcfTYear  @  Century +  year  !
;

: dcf.dbg.sec ( -- ) f.dcf.dbg fset? if dcf.dbg.second.over then ; \ debug
: dcf.dbg.min ( -- ) f.dcf.dbg fset? if dcf.dbg.minute.over then ; \ debug

: (dcf.sec)
  dcfClock timeup
  0 tick !
  
  \ evaluate pulse
  dcfPulse @ dcf.pulse>bit  dup dcfBit !
  ( bit ) case
    
    dcf.bit:0      of  0 dcfPos @ pos.cmd  dcf.dbg.sec  endof

    dcf.bit:1      of  1 dcfPos @ pos.cmd  dcf.dbg.sec  endof

    dcf.bit:error  of  f.dcf.err fset      dcf.dbg.sec  endof

    dcf.bit:sync59 of
      dcf.min
      dcf.tmp.counter.reset
      dcfErrCnt.set
    endof

  endcase
  dcfPulse.clr           \ reset pulse
  dcfPos++
;

: (dcf.min) ( -- )
  dcfPos @ #59 =  if dcfErrCnt-- then
  dcfErrCnt @ 0=  if f.dcf.err fclr then
  \ this block runs at the *end* of second "59" i.e.
  \ at the start of second "00", thus copy counters
  f.dcf.err fclr? if
    f.dcf.commit fset? if
      dcfTemp>dcfClock
      dcfClock>MasterClock
      space [char] C emit
      f.dcf.insync fset \ dcfClock is in sync!
      f.dcf.commit fclr
    then
  then
  dcf.dbg.sec
  dcf.dbg.min
  f.dcf.commit fclr? if
    \ clear debug flags --- this is only to demonstrate resync/commit
    f.dcf.dbg fclr
  then
;  

\ --- init: enable DCF clock -------------
: +dcf
  _dcf pin_input
  ['] (dcf.tick) to dcf.tick
  ['] (dcf.sec)  to dcf.sec
  ['] (dcf.min)  to dcf.min

  0 dcfPos !        \ assume we are at 0
  f.dcf.err fset    \ error unless proven otherwise
  f.dcf.commit fset \ request to set dcfClock
  f.dcf.insync fclr \ not in sync yet!
  dcfErrCnt.set
;
: -dcf
  ['] noop to dcf.tick
  ['] noop to dcf.sec
;

