\ 2017-09-11 ew
\
\ Written in 2017 by Erich Wälde <erich.waelde@forth-ev.de>
\
\ To the extent possible under law, the author(s) have dedicated
\ all copyright and related and neighboring rights to this software
\ to the public domain worldwide. This software is distributed
\ without any warranty.
\
\ You should have received a copy of the CC0 Public Domain
\ Dedication along with this software. If not, see
\ <http://creativecommons.org/publicdomain/zero/1.0/>.
\
\ constant
\     i2c_addr_rtc: $50
\ words:
\     rtc.get ( -- S/100 S M H d m Y )
\     rtc.set ( Y m d H M S S/100 -- )
\     rtc.show ( -- )

\ pcf8583:
\ addr
\ 0x00    control register 3: mask_flag
\ 0x01    sec/100.bcd
\ 0x02    sec.bcd
\ 0x03    min.bcd
\ 0x04    7: 0=24h clock, 6: am/pm flag
\         5-0: hour.bcd
\ 0x05    7-6: year%4 5-0: day.bcd
\ 0x06    7-5: weekdays unless maskbit,
\         4-0: month.bcd
\ static cmos ram:
\ 0x10,0x11 full year, not BCD

\ $50 constant i2c_addr_rtc


: rtc.get ( -- S/100 S M H d m Y )
    #6  #1 #1 i2c_addr_rtc i2c.m!n@     \ read 6 bytes start at addr 1
    #2 #16 #1 i2c_addr_rtc i2c.m!n@     \ read 2 bytes start at addr 16
    #8 lshift +                         \ convert to word YYYY

    (  year.dec )                 >r    \ from extra flash location
    ( month.bcd ) $1f and bcd>dec >r    \ dropped weekday?
    (   day.bcd ) $3f and bcd>dec >r    \ dropped year%4 ?
    (  hour.bcd ) $3f and bcd>dec >r    \ dropped 24 am/pm flags 
    (   min.bcd )         bcd>dec >r    \ 
    (   sec.bcd )         bcd>dec >r    \ 
    ( s/100.bcd )         bcd>dec       \ 
    r> r> r>  r> r> r>
;
: rtc.dot ( S/100 S M H d m Y -- )
    #4 u0.r [char] - emit               \ year
    #2 u0.r [char] - emit               \ month
    #2 u0.r [char] _ emit               \ day-of-month
    #2 u0.r [char] : emit               \ hour
    #2 u0.r [char] : emit               \ minute 
    #2 u0.r                             \ second
    drop                                \ s/100
;
: rtc.show ( -- )  rtc.get rtc.dot ;

: rtc.set.year ( year -- )
  dup >< swap                           \ -- y.h y.l
  $10 3 i2c_addr_rtc  i2c.n!            \ store at addr $10
;

: 2pick ( x2 x1 x0 -- x2 x1 x0 x2 )
    \ 2 pick
    >r over r> swap
;
: rtc.set ( Y m d H M S S/100 -- )
  ( s/100.dec ) dec>bcd >r
  (   sec.dec ) dec>bcd >r
  (   min.dec ) dec>bcd >r
  (  hour.dec ) dec>bcd >r              \ -- Y m d
  2pick #6 lshift +                     \ -- Y m Y%4|d 
  (   day.dec ) dec>bcd >r              \ -- Y m
  \ day-of-week bits \ fixme!!!
  ( month.dec ) dec>bcd
  r> r> r> r> r>                        \ -- Y m d H M S S/100
  $80 0 #8 i2c_addr_rtc i2c.n!          \ stop rtc send data
  $08 0 #2 i2c_addr_rtc i2c.n!          \ start rtc
  (  year.dec ) rtc.set.year            \ set year
;

  